@@ -6,8 +6,9 @@ import json
 from decimal import Decimal
 from typing import Generator, Mapping
 from urllib.parse import quote
+import requests
 
-from feldera.enums import PipelineFieldSelector
+from feldera.enums import BootstrapPolicy, PipelineFieldSelector, PipelineStatus
 from feldera.rest.config import Config
 from feldera.rest.feldera_config import FelderaConfig
 from feldera.rest.errors import FelderaTimeoutError, FelderaAPIError
@@ -110,24 +111,30 @@ class FelderaClient:
 
         return Pipeline.from_dict(resp)
 
-    def get_runtime_config(self, pipeline_name) -> dict:
+    def get_runtime_config(self, pipeline_name) -> Mapping[str, Any]:
         """
         Get the runtime config of a pipeline by name
 
         :param pipeline_name: The name of the pipeline
         """
 
-        resp: dict = self.http.get(f"/pipelines/{pipeline_name}")
+        resp: Mapping[str, Any] = self.http.get(f"/pipelines/{pipeline_name}")
 
-        return resp.get("runtime_config")
+        runtime_config: Mapping[str, Any] | None = resp.get("runtime_config")
+        if runtime_config is None:
+            raise ValueError(f"Pipeline {pipeline_name} has no runtime config")
 
-    def pipelines(self) -> list[Pipeline]:
+        return runtime_config
+
+    def pipelines(
+        self, selector: PipelineFieldSelector = PipelineFieldSelector.STATUS
+    ) -> list[Pipeline]:
         """
         Get all pipelines
         """
 
         resp = self.http.get(
-            path="/pipelines",
+            path=f"/pipelines?selector={selector.value}",
         )
 
         return [Pipeline.from_dict(pipeline) for pipeline in resp]
@@ -175,7 +182,7 @@ class FelderaClient:
         self,
         pipeline_name: str,
         state: str,
-        timeout_s: float = 300.0,
+        timeout_s: Optional[float] = None,
         start: bool = True,
     ):
         start_time = time.monotonic()
@@ -211,6 +218,46 @@ Reason: The pipeline is in a STOPPED state due to the following error:
             )
             time.sleep(0.1)
 
+    def __wait_for_pipeline_state_one_of(
+        self,
+        pipeline_name: str,
+        states: list[str],
+        timeout_s: float | None = None,
+        start: bool = True,
+    ) -> PipelineStatus:
+        start_time = time.monotonic()
+        states = [state.lower() for state in states]
+
+        while True:
+            if timeout_s is not None:
+                elapsed = time.monotonic() - start_time
+                if elapsed > timeout_s:
+                    raise TimeoutError(
+                        f"Timed out waiting for pipeline {pipeline_name} to"
+                        f"transition to one of the states: {states}"
+                    )
+
+            resp = self.get_pipeline(pipeline_name, PipelineFieldSelector.STATUS)
+            status = resp.deployment_status
+
+            if status.lower() in states:
+                return PipelineStatus.from_str(status)
+            elif (
+                status == "Stopped"
+                and len(resp.deployment_error or {}) > 0
+                and resp.deployment_desired_status == "Stopped"
+            ):
+                err_msg = "Unable to START the pipeline:\n" if start else ""
+                raise RuntimeError(
+                    f"""{err_msg}Unable to transition the pipeline to one of the states: {states}.
+Reason: The pipeline is in a STOPPED state due to the following error:
+{resp.deployment_error.get("message", "")}"""
+                )
+            logging.debug(
+                "still starting %s, waiting for 100 more milliseconds", pipeline_name
+            )
+            time.sleep(0.1)
+
     def create_pipeline(self, pipeline: Pipeline) -> Pipeline:
         """
         Create a pipeline if it doesn't exist and wait for it to compile
@@ -342,36 +389,36 @@ Reason: The pipeline is in a STOPPED state due to the following error:
                     yield chunk.decode("utf-8")
 
     def activate_pipeline(
-        self, pipeline_name: str, wait: bool = True, timeout_s: Optional[float] = 300
-    ):
+        self, pipeline_name: str, wait: bool = True, timeout_s: Optional[float] = None
+    ) -> Optional[PipelineStatus]:
         """
 
         :param pipeline_name: The name of the pipeline to activate
         :param wait: Set True to wait for the pipeline to activate. True by
             default
         :param timeout_s: The amount of time in seconds to wait for the
-            pipeline to activate. 300 seconds by default.
+            pipeline to activate.
         """
 
-        if timeout_s is None:
-            timeout_s = 300
-
         self.http.post(
             path=f"/pipelines/{pipeline_name}/activate",
         )
 
         if not wait:
-            return
+            return None
 
-        self.__wait_for_pipeline_state(pipeline_name, "running", timeout_s)
+        return self.__wait_for_pipeline_state_one_of(
+            pipeline_name, ["running", "AwaitingApproval"], timeout_s
+        )
 
     def _inner_start_pipeline(
         self,
         pipeline_name: str,
         initial: str = "running",
+        bootstrap_policy: Optional[BootstrapPolicy] = None,
         wait: bool = True,
-        timeout_s: Optional[float] = 300,
-    ):
+        timeout_s: Optional[float] = None,
+    ) -> Optional[PipelineStatus]:
         """
 
         :param pipeline_name: The name of the pipeline to start
@@ -379,67 +426,88 @@ Reason: The pipeline is in a STOPPED state due to the following error:
             by default.
         :param wait: Set True to wait for the pipeline to start. True by default
         :param timeout_s: The amount of time in seconds to wait for the
-            pipeline to start. 300 seconds by default.
+            pipeline to start.
         """
 
-        if timeout_s is None:
-            timeout_s = 300
+        params = {"initial": initial}
+        if bootstrap_policy is not None:
+            params["bootstrap_policy"] = bootstrap_policy.value
 
         self.http.post(
             path=f"/pipelines/{pipeline_name}/start",
-            params={"initial": initial},
+            params=params,
         )
 
         if not wait:
-            return
+            return None
 
-        self.__wait_for_pipeline_state(pipeline_name, initial, timeout_s)
+        return self.__wait_for_pipeline_state_one_of(
+            pipeline_name, [initial, "AwaitingApproval"], timeout_s
+        )
 
     def start_pipeline(
-        self, pipeline_name: str, wait: bool = True, timeout_s: Optional[float] = 300
-    ):
+        self,
+        pipeline_name: str,
+        bootstrap_policy: Optional[BootstrapPolicy] = None,
+        wait: bool = True,
+        timeout_s: Optional[float] = None,
+    ) -> Optional[PipelineStatus]:
         """
 
         :param pipeline_name: The name of the pipeline to start
         :param wait: Set True to wait for the pipeline to start.
             True by default
         :param timeout_s: The amount of time in seconds to wait for the
-            pipeline to start. 300 seconds by default.
+            pipeline to start.
         """
 
-        self._inner_start_pipeline(pipeline_name, "running", wait, timeout_s)
+        return self._inner_start_pipeline(
+            pipeline_name, "running", bootstrap_policy, wait, timeout_s
+        )
 
     def start_pipeline_as_paused(
-        self, pipeline_name: str, wait: bool = True, timeout_s: Optional[float] = 300
-    ):
+        self,
+        pipeline_name: str,
+        bootstrap_policy: Optional[BootstrapPolicy] = None,
+        wait: bool = True,
+        timeout_s: float | None = None,
+    ) -> Optional[PipelineStatus]:
         """
         :param pipeline_name: The name of the pipeline to start as paused.
         :param wait: Set True to wait for the pipeline to start as pause.
             True by default
         :param timeout_s: The amount of time in seconds to wait for the
-            pipeline to start. 300 seconds by default.
+            pipeline to start.
         """
 
-        self._inner_start_pipeline(pipeline_name, "paused", wait, timeout_s)
+        return self._inner_start_pipeline(
+            pipeline_name, "paused", bootstrap_policy, wait, timeout_s
+        )
 
     def start_pipeline_as_standby(
-        self, pipeline_name: str, wait: bool = True, timeout_s: Optional[float] = 300
+        self,
+        pipeline_name: str,
+        bootstrap_policy: Optional[BootstrapPolicy] = None,
+        wait: bool = True,
+        timeout_s: Optional[float] = None,
     ):
         """
         :param pipeline_name: The name of the pipeline to start as standby.
         :param wait: Set True to wait for the pipeline to start as standby.
             True by default
         :param timeout_s: The amount of time in seconds to wait for the
-            pipeline to start. 300 seconds by default.
+            pipeline to start.
         """
 
-        self._inner_start_pipeline(pipeline_name, "standby", wait, timeout_s)
+        self._inner_start_pipeline(
+            pipeline_name, "standby", bootstrap_policy, wait, timeout_s
+        )
 
     def resume_pipeline(
         self,
         pipeline_name: str,
         wait: bool = True,
-        timeout_s: Optional[float] = 300,
+        timeout_s: Optional[float] = None,
     ):
         """
         Resume a pipeline
@@ -447,12 +515,9 @@ Reason: The pipeline is in a STOPPED state due to the following error:
         :param pipeline_name: The name of the pipeline to stop
         :param wait: Set True to wait for the pipeline to pause. True by default
         :param timeout_s: The amount of time in seconds to wait for the pipeline
-            to pause. 300 seconds by default.
+            to pause.
         """
 
-        if timeout_s is None:
-            timeout_s = 300
-
         self.http.post(
             path=f"/pipelines/{pipeline_name}/resume",
         )
@@ -466,7 +531,7 @@ Reason: The pipeline is in a STOPPED state due to the following error:
         self,
         pipeline_name: str,
         wait: bool = True,
-        timeout_s: Optional[float] = 300,
+        timeout_s: Optional[float] = None,
     ):
         """
         Pause a pipeline
@@ -476,12 +541,9 @@ Reason: The pipeline is in a STOPPED state due to the following error:
              STOPPED state due to a failure.
         :param wait: Set True to wait for the pipeline to pause. True by default
         :param timeout_s: The amount of time in seconds to wait for the pipeline
-            to pause. 300 seconds by default.
+            to pause.
         """
 
-        if timeout_s is None:
-            timeout_s = 300
-
         self.http.post(
             path=f"/pipelines/{pipeline_name}/pause",
         )
@@ -491,12 +553,20 @@ Reason: The pipeline is in a STOPPED state due to the following error:
 
         self.__wait_for_pipeline_state(pipeline_name, "paused", timeout_s)
 
+    def approve_pipeline(
+        self,
+        pipeline_name: str,
+    ):
+        self.http.post(
+            path=f"/pipelines/{pipeline_name}/approve",
+        )
+
     def stop_pipeline(
         self,
         pipeline_name: str,
         force: bool,
         wait: bool = True,
-        timeout_s: Optional[float] = 300,
+        timeout_s: Optional[float] = None,
     ):
         """
         Stop a pipeline
@@ -506,12 +576,9 @@ Reason: The pipeline is in a STOPPED state due to the following error:
             Set False to automatically checkpoint before stopping.
         :param wait: Set True to wait for the pipeline to stop. True by default
         :param timeout_s: The amount of time in seconds to wait for the pipeline
-            to stop. Default is 300 seconds.
+            to stop.
         """
 
-        if timeout_s is None:
-            timeout_s = 300
-
         params = {"force": str(force).lower()}
 
         self.http.post(
@@ -524,7 +591,12 @@ Reason: The pipeline is in a STOPPED state due to the following error:
 
         start = time.monotonic()
 
-        while time.monotonic() - start < timeout_s:
+        while True:
+            if timeout_s is not None and time.monotonic() - start > timeout_s:
+                raise FelderaTimeoutError(
+                    f"timeout error: pipeline '{pipeline_name}' did not stop in {timeout_s} seconds"
+                )
+
             status = self.get_pipeline(
                 pipeline_name, PipelineFieldSelector.STATUS
             ).deployment_status
@@ -538,29 +610,26 @@ Reason: The pipeline is in a STOPPED state due to the following error:
             )
             time.sleep(0.1)
 
-        raise FelderaTimeoutError(
-            f"timeout error: pipeline '{pipeline_name}' did not stop in {timeout_s} seconds"
-        )
-
-    def clear_storage(self, pipeline_name: str, timeout_s: Optional[float] = 300):
+    def clear_storage(self, pipeline_name: str, timeout_s: Optional[float] = None):
         """
         Clears the storage from the pipeline.
         This operation cannot be canceled.
 
         :param pipeline_name: The name of the pipeline
         :param timeout_s: The amount of time in seconds to wait for the storage
-            to clear. Default is 300 seconds.
+            to clear.
         """
-        if timeout_s is None:
-            timeout_s = 300
-
         self.http.post(
             path=f"/pipelines/{pipeline_name}/clear",
         )
 
         start = time.monotonic()
 
-        while time.monotonic() - start < timeout_s:
+        while True:
+            if timeout_s is not None and time.monotonic() - start > timeout_s:
+                raise FelderaTimeoutError(
+                    f"timeout error: pipeline '{pipeline_name}' did not clear storage in {timeout_s} seconds"
+                )
             status = self.get_pipeline(
                 pipeline_name, PipelineFieldSelector.STATUS
             ).storage_status
@@ -574,10 +643,6 @@ Reason: The pipeline is in a STOPPED state due to the following error:
             )
             time.sleep(0.1)
 
-        raise FelderaTimeoutError(
-            f"timeout error: pipeline '{pipeline_name}' did not clear storage in {timeout_s} seconds"
-        )
-
     def start_transaction(self, pipeline_name: str) -> int:
         """
         Start a new transaction.
@@ -773,15 +838,15 @@ Reason: The pipeline is in a STOPPED state due to the following error:
                     _validate_no_none_keys_in_map(datum.get("insert", {}))
                     _validate_no_none_keys_in_map(datum.get("delete", {}))
             else:
-                data: dict = data
+                data: Mapping[str, Any] = data
                 _validate_no_none_keys_in_map(data.get("insert", {}))
                 _validate_no_none_keys_in_map(data.get("delete", {}))
         else:
             _validate_no_none_keys_in_map(data)
 
         # python sends `True` which isn't accepted by the backend
-        array = _prepare_boolean_input(array)
-        force = _prepare_boolean_input(force)
+        array: str = _prepare_boolean_input(array)
+        force: str = _prepare_boolean_input(force)
 
         params = {
             "force": force,
@@ -820,8 +885,37 @@ Reason: The pipeline is in a STOPPED state due to the following error:
 
         return token
 
+    def completion_token_processed(self, pipeline_name: str, token: str) -> bool:
+        """
+        Check whether the pipeline has finished processing all inputs received from the connector before
+        the token was generated.
+
+        :param pipeline_name: The name of the pipeline
+        :param token: The token to check for completion
+        :return: True if the pipeline has finished processing all inputs represented by the token, False otherwise
+        """
+
+        params = {
+            "token": token,
+        }
+
+        resp = self.http.get(
+            path=f"/pipelines/{quote(pipeline_name, safe='')}/completion_status",
+            params=params,
+        )
+
+        status: Optional[str] = resp.get("status")
+
+        if status is None:
+            raise FelderaAPIError(
+                f"got empty status when checking for completion status for token: {token}",
+                resp,
+            )
+
+        return status.lower() == "complete"
+
     def wait_for_token(
-        self, pipeline_name: str, token: str, timeout_s: Optional[float] = 600
+        self, pipeline_name: str, token: str, timeout_s: Optional[float] = None
     ):
         """
         Blocks until all records represented by this completion token have
@@ -830,13 +924,9 @@ Reason: The pipeline is in a STOPPED state due to the following error:
         :param pipeline_name: The name of the pipeline
         :param token: The token to check for completion
         :param timeout_s: The amount of time in seconds to wait for the pipeline
-            to process these records. Default 600s
+            to process these records.
         """
 
-        params = {
-            "token": token,
-        }
-
         start = time.monotonic()
         end = start + timeout_s if timeout_s else None
         initial_backoff = 0.1
@@ -849,22 +939,11 @@ Reason: The pipeline is in a STOPPED state due to the following error:
                 if time.monotonic() > end:
                     raise FelderaTimeoutError(
                         f"timeout error: pipeline '{pipeline_name}' did not"
-                        f" process records represented by token {token} within"
-                        f" {timeout_s}"
+                        + f" process records represented by token {token} within"
+                        + f" {timeout_s}"
                     )
 
-            resp = self.http.get(
-                path=f"/pipelines/{pipeline_name}/completion_status", params=params
-            )
-
-            status: Optional[str] = resp.get("status")
-            if status is None:
-                raise FelderaAPIError(
-                    f"got empty status when checking for completion status for token: {token}",
-                    resp,
-                )
-
-            if status.lower() == "complete":
+            if self.completion_token_processed(pipeline_name, token):
                 break
 
             elapsed = time.monotonic() - start
@@ -914,7 +993,7 @@ Reason: The pipeline is in a STOPPED state due to the following error:
 
         table_name = f'"{table_name}"' if case_sensitive else table_name
 
-        resp = self.http.post(
+        resp: requests.Response = self.http.post(
             path=f"/pipelines/{quote(pipeline_name, safe='')}/egress/{quote(table_name, safe='')}",
             params=params,
             stream=True,
@@ -1019,7 +1098,7 @@ Reason: The pipeline is in a STOPPED state due to the following error:
 
     def query_as_json(
         self, pipeline_name: str, query: str
-    ) -> Generator[dict, None, None]:
+    ) -> Generator[Mapping[str, Any], None, None]:
         """
         Executes an ad-hoc query on the specified pipeline and returns the result as a generator that yields
         rows of the query as Python dictionaries.
@@ -1128,3 +1207,31 @@ Reason: The pipeline is in a STOPPED state due to the following error:
                 buffer += chunk
 
         return buffer
+
+    def generate_completion_token(
+        self, pipeline_name: str, table_name: str, connector_name: str
+    ) -> str:
+        """
+        Generate a completion token that can be passed to :meth:`.FelderaClient.completion_token_processed` to
+        check whether the pipeline has finished processing all inputs received from the connector before
+        the token was generated.
+
+        :param pipeline_name: The name of the pipeline
+        :param table_name: The name of the table associated with this connector.
+        :param connector_name: The name of the connector.
+
+        :raises FelderaAPIError: If the connector cannot be found, or if the pipeline is not running.
+        """
+
+        resp = self.http.get(
+            path=f"/pipelines/{pipeline_name}/tables/{table_name}/connectors/{connector_name}/completion_token",
+        )
+
+        token: str | None = resp.get("token")
+
+        if token is None:
+            raise ValueError(
+                "got invalid response from feldera when generating completion token"
+            )
+
+        return token
